"""Base class for CFNgin hooks."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, Union, cast

from troposphere import Tags

from ...config.models.cfngin import CfnginStackDefinitionModel
from ...utils import BaseModel
from ..actions import deploy
from ..exceptions import StackFailed
from ..stack import Stack
from ..status import COMPLETE, FAILED, PENDING, SKIPPED, SUBMITTED
from .protocols import CfnginHookProtocol

if TYPE_CHECKING:
    from ..._logging import RunwayLogger
    from ...context import CfnginContext
    from ..blueprints.base import Blueprint
    from ..providers.aws.default import Provider
    from ..status import Status

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class HookArgsBaseModel(BaseModel):
    """Base model for hook args."""

    tags: Dict[str, str] = {}


class Hook(CfnginHookProtocol):
    """Base class for hooks.

    Not all hooks need to be classes and not all classes need to be hooks.

    Attributes:
        args: Keyword arguments passed to the hook, loaded into a MutableMap object.
        blueprint: Blueprint generated by the hook if it will be deploying a stack.
        context: Context instance. (passed in by CFNgin)
        provider Provider instance. (passed in by CFNgin)
        stack: Stack object if the hook deploys a stack.
        stack_name: Name of the stack created by the hook if a stack is to be created.

    """

    ARGS_PARSER: ClassVar[Type[HookArgsBaseModel]] = HookArgsBaseModel
    """Class used to parse arguments passed to the hook."""

    args: HookArgsBaseModel
    blueprint: Optional[Blueprint] = None
    context: CfnginContext
    provider: Provider
    stack: Optional[Stack] = None
    stack_name: str = "stack"

    def __init__(  # pylint: disable=super-init-not-called
        self, context: CfnginContext, provider: Provider, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            context: Context instance. (passed in by CFNgin)
            provider: Provider instance. (passed in by CFNgin)

        """
        kwargs.setdefault("tags", {})

        self.args = self.ARGS_PARSER.parse_obj(kwargs)
        self.args.tags.update(context.tags)
        self.context = context
        self.provider = provider
        # TODO BREAKING remove these from the primary base class
        self._deploy_action = HookDeployAction(self.context, self.provider)
        self._destroy_action = HookDestroyAction(self.context, self.provider)

    @property
    def tags(self) -> Tags:
        """Return tags that should be applied to any resource being created."""
        return Tags(**dict(self.context.tags, **self.args.tags))

    def generate_stack(self, **kwargs: Any) -> Stack:
        """Create a CFNgin Stack object."""
        definition = CfnginStackDefinitionModel.construct(
            name=self.stack_name, tags=self.args.tags, **kwargs
        )
        stack = Stack(definition, self.context)
        stack._blueprint = self.blueprint  # pylint: disable=protected-access
        return stack

    def get_template_description(self, suffix: Optional[str] = None) -> str:
        """Generate a template description.

        Args:
            suffix: Suffix to append to the end of a CloudFormation template
                description.

        """
        template = "Automatically generated by {}"
        if suffix:
            template += " - {}"
            return template.format(self.__class__.__module__, suffix)
        return template.format(self.__class__.__module__)

    def deploy_stack(self, stack: Optional[Stack] = None, wait: bool = False) -> Status:
        """Deploy a stack.

        Args:
            stack: A stack to act on.
            wait: Wither to wait for the stack to complete before returning.

        Returns:
            Ending status of the stack.

        """
        return self._run_stack_action(
            action=self._deploy_action, stack=stack, wait=wait
        )

    def destroy_stack(
        self, stack: Optional[Stack] = None, wait: bool = False
    ) -> Status:
        """Destroy a stack.

        Args:
            stack: A stack to act on.
            wait: Wither to wait for the stack to complete before returning.

        Returns:
            Ending status of the stack.

        """
        return self._run_stack_action(
            action=self._destroy_action, stack=stack, wait=wait
        )

    def post_deploy(self) -> Any:
        """Run during the **post_deploy** stage."""
        raise NotImplementedError

    def post_destroy(self) -> Any:
        """Run during the **post_destroy** stage."""
        raise NotImplementedError

    def pre_deploy(self) -> Any:
        """Run during the **pre_deploy** stage."""
        raise NotImplementedError

    def pre_destroy(self) -> Any:
        """Run during the **pre_destroy** stage."""
        raise NotImplementedError

    @staticmethod
    def _log_stack(stack: Stack, status: Status) -> None:
        """Log stack status. Mimics normal stack deployment.

        Args:
            stack: The stack being logged.
            status: The status being logged.

        """
        msg = f"{stack.name}:{status.name}"
        if status.reason:
            msg += f" ({status.reason})"
        if status.code == SUBMITTED.code:
            LOGGER.notice(msg)
        elif status.code == COMPLETE.code:
            LOGGER.success(msg)
        elif status.code == FAILED.code:
            LOGGER.error(msg)
        else:
            LOGGER.info(msg)

    def _run_stack_action(
        self,
        action: Union[HookDeployAction, HookDestroyAction],
        stack: Optional[Stack] = None,
        wait: bool = False,
    ) -> Status:
        """Run a CFNgin hook modified for use in hooks.

        Args:
            action: Action to be taken against a stack.
            stack: A stack to act on.
            wait: Wither to wait for the stack to complete before returning.

        Returns:
            Ending status of the stack.

        """
        stack = stack or self.stack
        if not stack:
            raise ValueError("stack required")
        status = action.run(stack=stack, status=PENDING)
        self._log_stack(stack, status)

        if wait and status != SKIPPED:
            status = self._wait_for_stack(
                action=action, stack=stack, last_status=status
            )
        return status

    def _wait_for_stack(
        self,
        action: Union[HookDeployAction, HookDestroyAction],
        last_status: Optional[Status] = None,
        stack: Optional[Stack] = None,
        till_reason: Optional[str] = None,
    ):
        """Wait for a CloudFormation stack to complete.

        Args:
            action: Action to be taken against a stack.
            last_status: The last status of the stack.
            stack: A stack that has been acted upon.
            till_reason: Status string to wait for before returning.
                ``COMPLETE`` or ``FAILED`` status will return before this
                condition if found.

        Returns:
            Ending status of the stack.

        Raises:
            StackFailed: Stack is in a failed state.

        """
        status = last_status or SUBMITTED
        stack = stack or self.stack
        if not stack:
            raise ValueError("stack required")

        while True:
            if status in (COMPLETE, FAILED):
                break
            if (till_reason and status.reason) and status.reason == till_reason:
                break
            if last_status and last_status.reason != status.reason:
                # log status changes like rollback
                self._log_stack(stack, status)
                last_status = status
            LOGGER.debug("waiting for stack to complete...")
            status = action.run(stack=stack, status=status)

        self._log_stack(stack, status)
        if status == FAILED:
            raise StackFailed(stack_name=stack.fqn, status_reason=status.reason)
        return status


# TODO BREAKING find a better place for this - can cause cyclic imports
class HookDeployAction(deploy.Action):
    """Deploy action that can be used from hooks."""

    def __init__(self, context: CfnginContext, provider: Provider):
        """Instantiate class.

        Args:
            context: The context for the current run.
            provider: The provider instance.

        """
        super().__init__(context)
        self._provider = provider

    @property
    def provider(self) -> Provider:
        """Override the inherited property to return the local provider."""
        return self._provider

    def build_provider(self) -> Provider:
        """Override the inherited method to always return local provider."""
        return self._provider

    def run(self, **kwargs: Any) -> Status:  # type: ignore
        """Run the action for one stack."""
        return self._launch_stack(**kwargs)


# the build action has logic to destroy stacks so we can just extend the
# HookDeployAction and change `run` in use the `_destroy_stack` method instead
class HookDestroyAction(HookDeployAction):
    """Destroy action that can be used from hooks."""

    def run(self, **kwargs: Any) -> Status:
        """Run the action for one stack."""
        return self._destroy_stack(**kwargs)



"""Docker logic for python."""
from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from docker.types.services import Mount

from .....compat import cached_property, shlex_join
from .....utils import Version
from ..docker import DockerDependencyInstaller

if TYPE_CHECKING:
    from docker.client import DockerClient

    from .....context import CfnginContext, RunwayContext
    from . import PythonProject


class PythonDockerDependencyInstaller(DockerDependencyInstaller):
    """Docker dependency installer for Python."""

    project: PythonProject

    def __init__(
        self,
        project: PythonProject,
        *,
        client: Optional[DockerClient] = None,
        context: Optional[Union[CfnginContext, RunwayContext]] = None,
    ) -> None:
        """Instantiate class.

        Args:
            project: awslambda project.
            client: Pre-configured :class:`docker.client.DockerClient`.
            context: CFNgin or Runway context object.

        """
        super().__init__(project, client=client, context=context)

    @cached_property
    def bind_mounts(self) -> List[Mount]:
        """Bind mounts that will be used by the container."""
        mounts = [*super().bind_mounts]
        if self.project.requirements_txt:
            mounts.append(
                Mount(
                    target=f"/var/task/{self.project.requirements_txt.name}",
                    source=str(self.project.requirements_txt),
                    type="bind",
                )
            )
        return mounts

    @cached_property
    def environmet_variables(self) -> Dict[str, str]:
        """Environment variables to pass to the docker container.

        This is a subset of the environment variables stored in the context
        object as some will cause issues if they are passed.

        """
        docker_env_vars = super().environmet_variables
        pip_env_vars = {
            k: v for k, v in self.ctx.env.vars.items() if k.startswith("PIP")
        }
        return {**docker_env_vars, **pip_env_vars}

    @cached_property
    def install_commands(self) -> List[str]:
        """Commands to run to install dependencies."""
        if self.project.requirements_txt:
            return [
                shlex_join(
                    self.project.pip.generate_install_command(
                        cache_dir=self.CACHE_DIR if self.project.cache_dir else None,
                        no_cache_dir=not self.project.args.use_cache,
                        no_deps=bool(self.project.poetry or self.project.pipenv),
                        requirements=f"/var/task/{self.project.requirements_txt.name}",
                        target=self.DEPENDENCY_DIR,
                    )
                    + (self.project.args.extend_pip_args or [])
                )
            ]
        return []

    @cached_property
    def python_version(self) -> Optional[Version]:
        """Version of Python installed in the docker container."""
        match = re.search(
            r"Python (?P<version>\S*)",
            "\n".join(self.run_command("python --version", level=logging.DEBUG)),
        )
        if not match:
            return None
        return Version(match.group("version"))

    @cached_property
    def runtime(self) -> Optional[str]:
        """AWS Lambda runtime determined from the docker container's Python version."""
        if not self.python_version:
            return None
        return f"python{self.python_version.major}.{self.python_version.minor}"



"""CFNgin Bucket Blueprint."""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Dict, Union

from troposphere import Equals, If, Not, NoValue, Or, Tag, Tags, s3

from ... import __version__
from ...compat import cached_property
from .base import Blueprint
from .variables.types import CFNString

if TYPE_CHECKING:
    from troposphere import AWSHelperFn

    from .type_defs import BlueprintVariableTypeDef


class CfnginBucket(Blueprint):
    """CFNgin Bucket Blueprint."""

    DESCRIPTION: ClassVar[str] = f"{__name__}.CFNginBucket (v{__version__})"
    VARIABLES: ClassVar[Dict[str, BlueprintVariableTypeDef]] = {
        "AccessControl": {
            "allowed_values": [
                "AuthenticatedRead",
                "AwsExecRead",
                "BucketOwnerFullControl",
                "BucketOwnerRead",
                "LogDeliveryWrite",
                "Private",
                "PublicRead",
                "PublicReadWrite",
            ],
            "default": "Private",
            "description": "A canned access control list (ACL) that grants "
            "predefined permissions to the bucket.",
            "type": CFNString,
        },
        "BucketName": {"description": "Name of the CFNgin bucket", "type": CFNString},
        "VersioningStatus": {
            "allowed_values": ["Enabled", "Suspended"],
            "default": "Enabled",
            "description": "Status of the bucket's VersioningConfiguration",
            "type": CFNString,
        },
    }

    @cached_property
    def bucket(self) -> s3.Bucket:
        """CFNgin Bucket."""
        bucket = s3.Bucket(
            "Bucket",
            AccessControl=self.variables["AccessControl"].ref,
            BucketEncryption=self.bucket_encryption,
            BucketName=self.bucket_name,
            OwnershipControls=s3.OwnershipControls(
                Rules=[s3.OwnershipControlsRule(ObjectOwnership="BucketOwnerPreferred")]
            ),
            Tags=self.bucket_tags,
            VersioningConfiguration=s3.VersioningConfiguration(
                Status=self.variables["VersioningStatus"].ref
            ),
        )
        self.add_output("BucketArn", bucket.get_att("Arn"))
        self.add_output("BucketDomainName", bucket.get_att("DomainName"))
        self.add_output("BucketName", bucket.ref())
        self.add_output(
            "BucketRegionalDomainName", bucket.get_att("RegionalDomainName")
        )
        return bucket

    @cached_property
    def bucket_encryption(self) -> Union[AWSHelperFn, s3.BucketEncryption]:
        """CFNgin bucket encryption.

        This cached property can be overridden in a subclass to customize the
        BucketEncryption property of the bucket without needing to override the
        bucket cached property.

        """
        return s3.BucketEncryption(
            ServerSideEncryptionConfiguration=[
                s3.ServerSideEncryptionRule(
                    ServerSideEncryptionByDefault=s3.ServerSideEncryptionByDefault(
                        SSEAlgorithm="AES256"
                    )
                )
            ]
        )

    @cached_property
    def bucket_name(self) -> AWSHelperFn:
        """CFNgin Bucket name."""
        condition = self.template.add_condition(
            "BucketNameProvided",
            Or(
                Not(Equals(self.variables["BucketName"].ref, "")),
                Not(Equals(self.variables["BucketName"].ref, "undefined")),
            ),
        )
        return If(condition, self.variables["BucketName"].ref, NoValue)

    @cached_property
    def bucket_tags(self) -> Tags:
        """CFNgin bucket tags.

        This cached property can be overridden in a subclass to customize the
        Tags property of the bucket without needing to override the bucket cached
        property.

        """
        return Tags(Tag("version", __version__))

    def create_template(self) -> None:
        """Create template."""
        self.template.set_description(self.DESCRIPTION)
        self.template.set_version("2010-09-09")
        self.template.add_resource(self.bucket)



"""AWS account."""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from ....compat import cached_property

if TYPE_CHECKING:
    import boto3

    from ....context import CfnginContext, RunwayContext


class AccountDetails:
    """AWS account details."""

    def __init__(self, context: Union[CfnginContext, RunwayContext]) -> None:
        """Instantiate class.

        Args:
            context: Runway context object.

        """
        self.__ctx = context

    @cached_property
    def aliases(self) -> List[str]:
        """Get the aliases of the AWS account."""
        # Super overkill here using pagination when an account can only
        # have a single alias, but at least this implementation should be
        # future-proof.
        aliases: List[str] = []
        paginator = self.__session.client("iam").get_paginator("list_account_aliases")
        response_iterator = paginator.paginate()
        for page in response_iterator:
            aliases.extend(page.get("AccountAliases", []))
        return aliases

    @cached_property
    def id(self) -> str:
        """Get the ID of the AWS account."""
        account_id = self.__session.client("sts").get_caller_identity().get("Account")
        if account_id:
            return account_id
        raise ValueError("get_caller_identity did not return Account")

    @cached_property
    def __session(self) -> boto3.Session:
        """Get a cached boto3 session.

        Session creation was moved out of class init to improve performance
        by only creating the session once it is needed.

        """
        return self.__ctx.get_session()



"""Runway Terraform Module options."""
# pylint: disable=no-self-argument
from __future__ import annotations

from typing import Dict, List, Optional, Union

from pydantic import Extra, Field, validator

from ...base import ConfigProperty


class RunwayTerraformArgsDataModel(ConfigProperty):
    """Modelf for Runway Terraform Module args option."""

    apply: List[str] = []
    init: List[str] = []
    plan: List[str] = []

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway Terraform Module args option"


class RunwayTerraformBackendConfigDataModel(ConfigProperty):
    """Model for Runway Terraform Module terraform_backend_config option."""

    bucket: Optional[str] = None
    dynamodb_table: Optional[str] = None
    region: Optional[str] = None
    workspace_key_prefix: Optional[str] = None

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway Terraform Module terraform_backend_config option"

    def __bool__(self) -> bool:
        """Evaluate the boolean value of the object instance."""
        data = self.dict(exclude_none=True)
        return "bucket" in data or "dynamodb_table" in data


class RunwayTerraformModuleOptionsDataModel(ConfigProperty):
    """Model for Runway Terraform Module options."""

    args: RunwayTerraformArgsDataModel = RunwayTerraformArgsDataModel()
    backend_config: RunwayTerraformBackendConfigDataModel = Field(
        default=RunwayTerraformBackendConfigDataModel(),
        alias="terraform_backend_config",
    )
    version: Optional[str] = Field(default=None, alias="terraform_version")
    workspace: Optional[str] = Field(default=None, alias="terraform_workspace")
    write_auto_tfvars: bool = Field(default=False, alias="terraform_write_auto_tfvars")

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.ignore
        title = "Runway Terraform Module options"

    @validator("args", pre=True)
    def _convert_args(
        cls, v: Union[List[str], Dict[str, List[str]]]
    ) -> Dict[str, List[str]]:
        """Convert args from list to dict."""
        if isinstance(v, list):
            return {"apply": v}
        return v



"""CFNgin init action."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from ...compat import cached_property
from ...config.models.cfngin import CfnginStackDefinitionModel
from ...core.providers.aws.s3 import Bucket
from ..exceptions import CfnginBucketAccessDenied
from . import deploy
from .base import BaseAction

if TYPE_CHECKING:
    import threading

    from ..._logging import RunwayLogger
    from ...context import CfnginContext
    from ..providers.aws.default import ProviderBuilder

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class Action(BaseAction):
    """Initialize environment."""

    NAME = "init"
    DESCRIPTION = "Initialize environment"

    def __init__(
        self,
        context: CfnginContext,
        provider_builder: Optional[ProviderBuilder] = None,
        cancel: Optional[threading.Event] = None,
    ):
        """Instantiate class.

        This class creates a copy of the context object prior to initialization
        as some of it can perform destructive actions on the context object.

        Args:
            context: The context for the current run.
            provider_builder: An object that will build a provider that will be
                interacted with in order to perform the necessary actions.
            cancel: Cancel handler.

        """
        super().__init__(
            context=context.copy(), provider_builder=provider_builder, cancel=cancel
        )

    @property
    def _stack_action(self) -> Any:
        """Run against a step."""
        return None

    @cached_property
    def cfngin_bucket(self) -> Optional[Bucket]:
        """CFNgin bucket.

        Raises:
            CfnginBucketRequired: cfngin_bucket not defined.

        """
        if not self.context.bucket_name:
            return None
        return Bucket(
            self.context,
            name=self.context.bucket_name,
            region=self.context.bucket_region,
        )

    @cached_property
    def default_cfngin_bucket_stack(self) -> CfnginStackDefinitionModel:
        """CFNgin bucket stack."""
        return CfnginStackDefinitionModel(
            class_path="runway.cfngin.blueprints.cfngin_bucket.CfnginBucket",
            in_progress_behavior="wait",
            name="cfngin-bucket",
            termination_protection=True,
            variables={"BucketName": self.context.bucket_name},
        )

    def run(
        self,
        *,
        concurrency: int = 0,
        dump: Union[bool, str] = False,  # pylint: disable=unused-argument
        force: bool = False,  # pylint: disable=unused-argument
        outline: bool = False,  # pylint: disable=unused-argument
        tail: bool = False,
        upload_disabled: bool = True,  # pylint: disable=unused-argument
        **_kwargs: Any,
    ) -> None:
        """Run the action.

        Args:
            concurrency: The maximum number of concurrent deployments.
            dump: Not used by this action
            force: Not used by this action.
            outline: Not used by this action.
            tail: Tail the stack's events.
            upload_disabled: Not used by this action.

        Raises:
            CfnginBucketAccessDenied: Could not head cfngin_bucket.

        """
        # attr can be falsy but still be Bucket so need to check type
        if not isinstance(self.cfngin_bucket, Bucket):
            LOGGER.info("skipped; cfngin_bucket not defined")
            return
        if self.cfngin_bucket.forbidden:
            raise CfnginBucketAccessDenied(bucket_name=self.cfngin_bucket.name)
        if self.cfngin_bucket.exists:
            LOGGER.info("cfngin_bucket %s already exists", self.cfngin_bucket.name)
            return
        if self.context.get_stack("cfngin-bucket"):
            LOGGER.verbose(
                "found stack for creating cfngin_bucket: cfngin-bucket",
            )
            self.context.stack_names = ["cfngin-bucket"]
        else:
            LOGGER.notice("using default blueprint to create cfngin_bucket...")
            self.context.config.stacks = [self.default_cfngin_bucket_stack]
            # clear cached values that were populated by checking the previous condition
            self.context._del_cached_property(  # pylint: disable=protected-access
                "stacks", "stacks_dict"
            )
        if self.provider_builder:
            self.provider_builder.region = self.context.bucket_region
        deploy.Action(
            context=self.context,
            provider_builder=self.provider_builder,
            cancel=self.cancel,
        ).run(
            concurrency=concurrency,
            tail=tail,
            upload_disabled=True,
        )
        return

    def pre_run(
        self,
        *,
        dump: Union[bool, str] = False,
        outline: bool = False,
        **__kwargs: Any,
    ) -> None:
        """Do nothing."""

    def post_run(
        self,
        *,
        dump: Union[bool, str] = False,
        outline: bool = False,
        **__kwargs: Any,
    ) -> None:
        """Do nothing."""



"""Runway config components."""
from ._deployment_def import RunwayDeploymentDefinition
from ._module_def import RunwayModuleDefinition
from ._test_def import (
    CfnLintRunwayTestDefinition,
    RunwayTestDefinition,
    ScriptRunwayTestDefinition,
    YamlLintRunwayTestDefinition,
)
from ._variables_def import RunwayVariablesDefinition

__all__ = [
    "CfnLintRunwayTestDefinition",
    "RunwayDeploymentDefinition",
    "RunwayModuleDefinition",
    "RunwayTestDefinition",
    "RunwayVariablesDefinition",
    "ScriptRunwayTestDefinition",
    "YamlLintRunwayTestDefinition",
]



"""Handle python requirements."""
from ._deployment_package import PythonDeploymentPackage
from ._docker import PythonDockerDependencyInstaller
from ._project import PythonProject

__all__ = [
    "PythonDeploymentPackage",
    "PythonDockerDependencyInstaller",
    "PythonProject",
]



"""CFNgin stack."""
from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, cast

from typing_extensions import Literal

from runway.utils import load_object_from_string
from runway.variables import Variable, resolve_variables

from .blueprints.raw import RawTemplateBlueprint

if TYPE_CHECKING:
    from ..config.models.cfngin import CfnginStackDefinitionModel
    from ..context import CfnginContext
    from .blueprints.base import Blueprint
    from .providers.aws.default import Provider


def _initialize_variables(
    stack_def: CfnginStackDefinitionModel, variables: Optional[Dict[str, Any]] = None
) -> List[Variable]:
    """Convert defined variables into a list of ``Variable`` for consumption.

    Args:
        stack_def: The stack definition being worked on.
        variables: Optional, explicit variables.

    Returns:
        Contains key/value pairs of the collected variables.

    Raises:
        AttributeError: Raised when the stack definition contains an invalid
            attribute. Currently only when using old parameters, rather than
            variables.

    """
    variables = variables or stack_def.variables or {}
    variable_values = deepcopy(variables)
    return [Variable(k, v, "cfngin") for k, v in variable_values.items()]


class Stack:
    """Represents gathered information about a stack to be built/updated.

    Attributes:
        definition: The stack definition from the config.
        enabled: Whether this stack is enabled
        force: Whether to force updates on this stack.
        fqn: Fully qualified name of the stack. Combines the stack name
            and current namespace.
        in_progress_behavior: The behavior for when a stack is in
            ``CREATE_IN_PROGRESS`` or ``UPDATE_IN_PROGRESS``.
        locked: Whether or not the stack is locked.
        logging: Whether logging is enabled.
        mappings: Cloudformation mappings passed to the blueprint.
        name: Name of the stack taken from the definition.
        outputs: CloudFormation Stack outputs.
        protected: Whether this stack is protected.
        termination_protection: The state of termination protection
            to apply to the stack.
        variables: Variables for the stack.

    """

    _blueprint: Optional[Blueprint]
    _stack_policy: Optional[str]

    context: CfnginContext
    definition: CfnginStackDefinitionModel
    enabled: bool
    force: bool
    fqn: str
    in_progress_behavior: Optional[Literal["wait"]]
    locked: bool
    logging: bool
    mappings: Dict[str, Dict[str, Dict[str, Any]]]
    name: str
    outputs: Dict[str, Any]
    protected: bool
    termination_protection: bool
    variables: List[Variable]

    def __init__(
        self,
        definition: CfnginStackDefinitionModel,
        context: CfnginContext,
        *,
        variables: Optional[Dict[str, Any]] = None,
        mappings: Dict[str, Dict[str, Dict[str, Any]]] = None,
        locked: bool = False,
        force: bool = False,
        enabled: bool = True,
        protected: bool = False,
    ):
        """Instantiate class.

        Args:
            definition: A stack definition.
            context: Current context for deploying the stack.
            variables: Variables for the stack.
            mappings: Cloudformation mappings passed to the blueprint.
            locked: Whether or not the stack is locked.
            force: Whether to force updates on this stack.
            enabled: Whether this stack is enabled
            protected: Whether this stack is protected.

        """
        self._blueprint = None
        self._stack_policy = None

        self.name = definition.name  # dependency of other attrs
        self.context = context
        self.definition = definition
        self.enabled = enabled
        self.force = force
        self.fqn = context.get_fqn(definition.stack_name or self.name)
        self.in_progress_behavior = definition.in_progress_behavior
        self.locked = locked
        self.logging = True
        self.mappings = mappings or {}
        self.outputs = {}
        self.protected = protected
        self.termination_protection = definition.termination_protection
        self.variables = _initialize_variables(definition, variables)

    @property
    def required_by(self) -> Set[str]:
        """Return a list of stack names that depend on this stack."""
        return set(self.definition.required_by)

    @property
    def requires(self) -> Set[str]:
        """Return a list of stack names this stack depends on."""
        requires = set(self.definition.requires or [])

        # Add any dependencies based on output lookups
        for variable in self.variables:
            deps = variable.dependencies
            if self.name in deps:
                raise ValueError(
                    f"Variable {variable.name} in stack {self.name} has a circular reference"
                )
            requires.update(deps)
        return requires

    @property
    def stack_policy(self) -> Optional[str]:
        """Return the Stack Policy to use for this stack."""
        if not self._stack_policy:
            self._stack_policy = None
            if self.definition.stack_policy_path:
                with open(self.definition.stack_policy_path, encoding="utf-8") as file_:
                    self._stack_policy = file_.read()

        return self._stack_policy

    @property
    def blueprint(self) -> Blueprint:
        """Return the blueprint associated with this stack."""
        if not self._blueprint:
            kwargs: Dict[str, Any] = {}
            if self.definition.class_path:
                class_path = self.definition.class_path
                blueprint_class = load_object_from_string(class_path)
                if not hasattr(blueprint_class, "rendered"):
                    raise AttributeError(
                        f'Stack class {class_path} does not have a "rendered" attribute.'
                    )
            elif self.definition.template_path:
                blueprint_class = RawTemplateBlueprint
                kwargs["raw_template_path"] = self.definition.template_path
            else:
                raise AttributeError(
                    "Stack does not have a defined class or template path."
                )

            self._blueprint = cast(
                "Blueprint",
                blueprint_class(
                    name=self.name,
                    context=self.context,
                    mappings=self.mappings,
                    description=self.definition.description,
                    **kwargs,
                ),
            )
        return self._blueprint

    @property
    def tags(self) -> Dict[str, Any]:
        """Return the tags that should be set on this stack.

        Includes both the global tags, as well as any stack specific tags
        or overrides.

        """
        tags = self.definition.tags or {}
        return dict(self.context.tags, **tags)

    @property
    def parameter_values(self) -> Dict[str, Any]:
        """Return all CloudFormation Parameters for the stack.

        CloudFormation Parameters can be specified via Blueprint Variables
        with a :class:`runway.cfngin.blueprints.variables.types.CFNType`
        ``type``.

        Returns:
            Dictionary of ``<parameter name>: <parameter value>``.

        """
        return self.blueprint.parameter_values

    @property
    def all_parameter_definitions(self) -> Dict[str, Any]:
        """Return all parameters in the blueprint/template."""
        return self.blueprint.parameter_definitions

    @property
    def required_parameter_definitions(self) -> Dict[str, Any]:
        """Return all CloudFormation Parameters without a default value."""
        return self.blueprint.required_parameter_definitions

    def resolve(
        self, context: CfnginContext, provider: Optional[Provider] = None
    ) -> None:
        """Resolve the Stack variables.

        This resolves the Stack variables and then prepares the Blueprint for
        rendering by passing the resolved variables to the Blueprint.

        Args:
            context: CFNgin context.
            provider: Subclass of the base provider.

        """
        resolve_variables(self.variables, context, provider)
        self.blueprint.resolve_variables(self.variables)

    def set_outputs(self, outputs: Dict[str, Any]) -> None:
        """Set stack outputs to the provided value.

        Args:
            outputs: CloudFormation Stack outputs.

        """
        self.outputs = outputs

    def __repr__(self) -> str:
        """Object represented as a string."""
        return self.fqn



"""Base classes for dependency managers."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Tuple, Union

from ..compat import cached_property
from ..mixins import CliInterfaceMixin

if TYPE_CHECKING:
    from _typeshed import StrPath

    from ..context import CfnginContext, RunwayContext
    from ..utils import Version


class DependencyManager(CliInterfaceMixin):
    """Dependency manager for the AWS Lambda runtime.

    Dependency managers are interfaced with via subprocess to ensure that the
    correct version is being used. This is primarily target at Python
    dependency manager that we could import and use directly.

    """

    CONFIG_FILES: ClassVar[Tuple[str, ...]]
    """Configuration files used by the dependency manager."""

    def __init__(
        self, context: Union[CfnginContext, RunwayContext], cwd: StrPath
    ) -> None:
        """Instantiate class.

        Args:
            context: CFNgin or Runway context object.
            cwd: Working directory where commands will be run.

        """
        self.ctx = context
        self.cwd = cwd if isinstance(cwd, Path) else Path(cwd)

    @cached_property
    def version(self) -> Version:
        """Get executable version."""
        raise NotImplementedError

    @classmethod
    def dir_is_project(cls, directory: StrPath, **__kwargs: Any) -> bool:
        """Determine if the directory contains a project for this dependency manager.

        Args:
            directory: Directory to check.

        """
        raise NotImplementedError



"""Handler for fetching outputs from a stack in the current namespace."""
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Tuple

from typing_extensions import Final, Literal

from ....lookups.handlers.base import LookupHandler
from ....lookups.handlers.cfn import CfnLookup
from ....utils import DOC_SITE
from .output import OutputQuery, deconstruct

if TYPE_CHECKING:
    from ....context import CfnginContext
    from ...providers.aws.default import Provider

LOGGER = logging.getLogger(__name__)


class RxrefLookup(LookupHandler):
    """Rxref lookup."""

    DEPRECATION_MSG = (
        'lookup query syntax "<relative-stack-name>::<OutputName>" has been deprecated; '
        "to learn how to use the new lookup query syntax visit "
        f"{DOC_SITE}/page/cfngin/lookups/rxref.html"
    )
    TYPE_NAME: Final[Literal["rxref"]] = "rxref"
    """Name that the Lookup is registered as."""

    @classmethod
    def legacy_parse(cls, value: str) -> Tuple[OutputQuery, Dict[str, str]]:
        """Retain support for legacy lookup syntax.

        Format of value:
            <relative-stack-name>::<OutputName>

        """
        LOGGER.warning("${%s %s}: %s", cls.TYPE_NAME, value, cls.DEPRECATION_MSG)
        return deconstruct(value), {}

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls, value: str, context: CfnginContext, provider: Provider, **_: Any
    ) -> Any:
        """Fetch an output from the designated stack in the current namespace.

        The ``output`` lookup supports fetching outputs from stacks created
        within a single config file. Sometimes it's useful to fetch outputs
        from stacks created outside of the current config file but using the
        same namespace. ``rxref`` supports this by using the
        :class:`runway.context.CfnginContext` to expand the fqn of the stack.

        Args:
            value: Parameter(s) given to this lookup. `"<relative-stack-name>.<OutputName>``
            context: Context instance.
            provider: Provider instance.

        """
        try:
            raw_query, _args = cls.parse(value)
            query = OutputQuery(*raw_query.split("."))
            colon_split = value.split("::", 1)
            raw_args = colon_split[1] if len(colon_split) > 1 else ""
        except ValueError:
            query, _args = cls.legacy_parse(value)
            raw_args = ""
        stack_fqn = context.get_fqn(query.stack_name)
        return CfnLookup.handle(
            f"{stack_fqn}.{query.output_name}" + (f"::{raw_args}" if raw_args else ""),
            context=context,
            provider=provider,
        )
